#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include "packet.h"
#include "config.h"
#include "value.h"
#include "lists.h"
#include "i3.h"

long	 i3_ntypes;

void packet_init (void)
	{ i3_ntypes = sizeof(i3_typedata) / sizeof(i3_typedata[0]); }

char * packet_int_to_desc (int p_type)
	{ return ((p_type<i3_ntypes)?i3_typedata[p_type].i3_desc:NULL); }
int
packet_desc_to_int (char *p_desc)
{
	int		  i = 0;
	while (i3_typedata[i].i3_desc &&
			strcmp(i3_typedata[i].i3_desc, p_desc)) i++;
	return (i3_typedata[i].i3_desc?i:-1);
}

listentry *
packet_node_new (int v_type, void *v_data)
{
	listentry	*l;
	value		*v;
	v = value_new(v_type);
	switch(v_type) {
		case MM_STRING:
			v->v_string = (v_data?strdup((char *)(v_data)):NULL);
			v->v_value = (v_data?strlen((char *)(v_data)):0);
			break;
		case MM_VALUE:
			v->v_value = (long)(v_data);
			break;
		case MM_ARRAY:
			v->v_array = v_data;
			break;
		default:
			fprintf(stderr, "packet_node_new called with %d\n", v_type);
	}
	l = list_item_new();
	l->data = (void *)(v);
	return l;
}

#define ST_NAMES char *s_uname, char *t_mname, char *t_uname
#define L_NSTR(r, s) list_append(r, packet_node_new(MM_STRING, s))
#define L_NVAL(r, v) list_append(r, packet_node_new(MM_VALUE, (void*)v))
#define L_NARR(r, d) list_append(r, packet_node_new(MM_ARRAY, d))

void
packet_head_create (listentry *root, char *p_desc, ST_NAMES)
{
	if (packet_desc_to_int(p_desc) == -1) return;

	L_NSTR(root, p_desc);
	L_NVAL(root, 5);
	L_NSTR(root, CONFIG_MUDNAME);
	L_NSTR(root, s_uname);
	L_NSTR(root, t_mname);
	L_NSTR(root, t_uname);
}

void
packet_tell_create (listentry *root, ST_NAMES,
		char *s_vname, char *msg)
{
	packet_head_create(root, "tell", s_uname, t_mname, t_uname);
	L_NSTR(root, s_vname);
	L_NSTR(root, msg);
}

void
packet_emoteto_create (listentry *root, ST_NAMES,
		char *s_vname, char *msg)
{
	packet_head_create(root, "emoteto", s_uname, t_mname, t_uname);
	L_NSTR(root, s_vname);
	L_NSTR(root, msg);
}

void
packet_who_req_create (listentry *root, ST_NAMES)
{
	packet_head_create(root, "who-req", s_uname, t_mname, NULL);
}

void
packet_who_reply_create (listentry *root, ST_NAMES, void *who_data)
{
	packet_head_create(root, "who-reply", NULL, t_mname, t_uname);
	L_NARR(root, who_data);
}

void
packet_finger_req_create (listentry *root, ST_NAMES, char *f_uname)
{
	packet_head_create(root, "finger-req", s_uname, t_mname, NULL);
	L_NSTR(root, f_uname);
}

void
packet_finger_reply_create (listentry *root, ST_NAMES,
		char *f_vname, char *f_title, char *f_rname, char *f_email,
		char *f_times, long f_idle, char *f_ip, char *f_level,
		char *f_extra)
{
	packet_head_create(root, "finger-reply", NULL, t_mname, t_uname);
	L_NSTR(root, f_vname);
	L_NSTR(root, f_title);
	L_NSTR(root, f_rname);
	L_NSTR(root, f_email);
	L_NSTR(root, f_times);
	L_NVAL(root, f_idle);
	L_NSTR(root, f_ip);
	L_NSTR(root, f_level);
	L_NSTR(root, f_extra);
}

void
packet_locate_req_create (listentry *root, ST_NAMES, char *l_uname)
{
	packet_head_create(root, "locate-req", s_uname, NULL, NULL);
	L_NSTR(root, l_uname);
}

void
packet_locate_reply_create (listentry *root, ST_NAMES,
		char *l_mname, char *l_vname, long l_idle, char *l_status)
{
	packet_head_create(root, "locate-reply", NULL, t_mname, t_uname);
	L_NSTR(root, l_mname);
	L_NSTR(root, l_vname);
	L_NVAL(root, l_idle);
	L_NSTR(root, l_status);
}

/* DON'T FUCK THIS UP */
void
packet_generic_create (listentry *root, int p_type, 
		char *s_uname, char *t_mname, char *t_uname, ... )
{
	va_list	 ap;
	char	*p_desc;
	char	*p_fmt;

	p_desc = i3_typedata[p_type].i3_desc;
	p_fmt = i3_typedata[p_type].i3_fmt;

	va_start(ap, t_uname);
	
	packet_head_create(root, "locate-reply", s_uname, t_mname, t_uname);

	while (*p_fmt) {
		switch(*p_fmt++) {
			case 's':
				L_NSTR(root, va_arg(ap, char *));
				break;
			case 'v':
				L_NVAL(root, va_arg(ap, long));
				break;
			case 'a':
				L_NARR(root, va_arg(ap, void *));
				break;
			default:
				fprintf(stderr, "Error in format string.");
		}
	}
	va_end(ap);
}
