#include <stdlib.h>
#include <string.h>
#include "lists.h"

inline
listentry *
list_item_new (void)
{
	listentry	*node;

	node = (listentry *)malloc(sizeof(listentry));

#ifdef CONFIG_DEBUG
	node->next = NULL;
	node->prev = NULL;
#endif
	node->label = NULL;
	node->data = NULL;

	return node;
}

listentry *
list_new (void)
{
	listentry	*node;

	node = list_item_new();
	node->prev = node;
	node->next = node;
	return node;
}

void
list_append (listentry *rootnode, listentry *node)
{
	node->next = rootnode;
	node->prev = rootnode->prev;

	rootnode->prev->next = node;
	rootnode->prev = node;
}

void
list_prepend (listentry *rootnode, listentry *node)
{
	list_append (rootnode->next, node);
}

void
/* This does not zero the other pointers as it is likely that this
 * list item is just about to be added in elsewhere. */
list_remove (listentry *node)
{
	node->next->prev = node->prev;
	node->prev->next = node->next;
}

long
list_length (listentry *root)
{
	long		 i = 0;
	listentry	*tempnode;
	tempnode = root->next;
	while (tempnode != root) {
		i++;
		tempnode = tempnode->next;
	}
	return i;
}

#if 0
int
list_data_value_compare (listentry *node1, listentry *node2)
{
	return (((long)(node2->data)) - ((long)(node1->data)));
}
#endif

int
list_data_rcompare (listentry *node1, listentry *node2)
{
	return (0 - list_data_compare(node1, node2));
}

int
list_data_compare (listentry *node1, listentry *node2)
{
	return (((long)(node2->data)) - ((long)(node1->data)));
}

int
list_label_compare (listentry *node1, listentry *node2)
{
	return strcmp(node2->label, node1->label);
}

void
list_qsort_segment (listentry *startnode, listentry *endnode,
				int (*compare)(listentry *node1, listentry *node2))
{
	listentry	*centrenode;
	listentry	*tempnode;
	listentry	*movenode;

	if (startnode->next == endnode)
		return;
	else {
		centrenode = startnode->next;
		tempnode = centrenode->next;
		while (tempnode != endnode) {
			if (compare(tempnode, centrenode) > 0) {
				movenode = tempnode;
				tempnode = tempnode->next;
				list_remove (movenode);
				list_append (centrenode, movenode);
			}
			else
				tempnode = tempnode->next;
		}
		list_qsort_segment (startnode, centrenode, compare);
		list_qsort_segment (centrenode, endnode, compare);
	}
}

inline void
list_qsort (listentry *rootnode, int compare())
{
	list_qsort_segment (rootnode, rootnode, compare);
}

inline void
list_sort (listentry *rootnode)
{
	list_qsort(rootnode, list_data_compare);
}

/* XXX Optimise these if statements */
listentry *
list_msort_merge (listentry *root1, listentry *root2, int compare())
{
	listentry	*list;
	listentry	*temp;
	listentry	*node1;
	listentry	*node2;
	list = list_new();
	node1 = root1->next;
	node2 = root2->next;

	while (node1 != root1) {
		while ((node2 != root2) && (compare(node1, node2) > 0))
		{
			temp = node2;
			node2 = node2->next;
			list_remove(temp);
			list_append(list, temp);
		}
		temp = node1;
		node1 = node1->next;
		list_remove(temp);
		list_append(list, temp);
	}
	list_combine(list, root2);
	list_destroy(root1);
	return list;
}

#if 0
listentry *
list_msort_merge_special_add (listentry *root1, listentry *root2, int compare())
{
	listentry	*list;
	listentry	*temp;
	listentry	*node1;
	listentry	*node2;
	int			 value;

	list = list_new();
	list->data = (void *)
					((long)(root1->data) +
					 (long)(root2->data));
	node1 = root1->next;
	node2 = root2->next;

	while ((node1 != root1) && (node2 != root2))
	{
		value = compare(node1, node2);
		if (value > 0) {
			temp = node2;
			node2 = node2->next;
			list_remove(temp);
			list_append(list, temp);
		}
		else if (value == 0) {
			temp = node2;
			temp->data = (void *)
				((long)(node1->data) + (long)(node2->data));
			node2 = node2->next;
			list_remove(temp);
			list_append(list, temp);

			temp = node1;
			node1 = node1->next;
			list_remove(temp);
		}
		else {
			temp = node1;
			node1 = node1->next;
			list_remove(temp);
			list_append(list, temp);
		}
	}

	if (node2 != root2) {
		list_combine(list, root2);
		list_destroy(root1);
	}
	else if (node1 != root1) {
		list_combine(list, root1);
		list_destroy(root2);
	}
	return list;
}

listentry *
list_msort_merge_special_combine (listentry *root1, listentry *root2, int compare())
{
	listentry	*list;
	listentry	*temp;
	listentry	*node1;
	listentry	*node2;
	int			 value;

	list = list_new();
	list->data = (void *)
					((long)(root1->data) +
					 (long)(root2->data));
	node1 = root1->next;
	node2 = root2->next;

	while ((node1 != root1) && (node2 != root2))
	{
		value = compare(node1, node2);
		if (value > 0) {
			temp = node2;
			node2 = node2->next;
			list_remove(temp);
			list_append(list, temp);
		}
		else if (value == 0) {
			temp = node2;
			if (!((long)(node1->data) == (long)(node2->data))) {
				fprintf(stderr, "ERROR: Values in list_msort_merge_special_combine not equal\n");
			}
			node2 = node2->next;
			list_remove(temp);
			list_append(list, temp);

			temp = node1;
			node1 = node1->next;
			list_remove(temp);
		}
		else {
			temp = node1;
			node1 = node1->next;
			list_remove(temp);
			list_append(list, temp);
		}
	}

	if (node2 != root2) {
		list_combine(list, root2);
		list_destroy(root1);
	}
	else if (node1 != root1) {
		list_combine(list, root1);
		list_destroy(root2);
	}
	return list;
}
#endif

void
list_item_destroy_data (listentry *tempnode, int flags)
{
	if (tempnode->data)
		if (flags & L_FDATA)
			free (tempnode->data);

	if (tempnode->label)
		if (flags & L_FLABEL)
			free (tempnode->label);
#ifdef CONFIG_DEBUG
		else
			printf("Label %s unfreed\n", tempnode->label);
#endif
}

void
list_item_destroy (listentry *tempnode, int flags)
{
	list_item_destroy_data (tempnode, flags);
	free(tempnode);
}

/* Note, rootnode never has a malloc()ed label. */
/* No longer true. Now if it has a label, it is malloc()ed. */
void
list_destroy (listentry *rootnode)
{
	listentry	*tempnode;

	tempnode = rootnode->next;
	while (tempnode != rootnode) {
		tempnode = tempnode->next;
#if 0
		list_item_destroy_data (tempnode, (int)(rootnode->data));
		free (tempnode->prev);
#endif
		list_item_destroy (tempnode->prev, (int)(rootnode->data));
	}
#if 0
	if (rootnode->data)
		fprintf(stderr, "ROOTNODE->data NOT FREED, value %p\n",
					rootnode->data);
#endif
	/* This was commented but should now be OK. */
	if ((int)(rootnode->data) & L_FRLABEL)
		if (rootnode->label)
			free(rootnode->label);
	free (rootnode);
}

/* Do not call on lists with labels undefined. For speed, it is not
 * checked. */
long
list_max_label (listentry *rootnode)
{
	listentry	*tempnode;
	long		 max = 0;

	tempnode = rootnode->next;
	while (tempnode != rootnode) {
		if (strlen(tempnode->label) > max)
			max = strlen(tempnode->label);
		tempnode = tempnode->next;
	}
	return max;
}

void
list_combine (listentry *list1, listentry *list2)
{
	if (list2->next != list2) {
		list2->prev->next = list1;
		list1->prev->next = list2->next;
		list2->next->prev = list1->prev;
		list1->prev = list2->prev;

		list2->next = list2;
		list2->prev = list2;
	}
	list_destroy(list2);		/* This catches all the L_F* */
}
