%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "value.h"
#include "mudmode.h"
#include "lists.h"
#include "tree.h"
%}

%option noyywrap
%option always-interactive
%pointer

%x ERROR

dDIGIT	[[:digit:]]
dCHAR	[[:alpha:]]
dWHITE	[[:blank:]]
dMUDCHAR	[^\n\"]
dSTRING		\"{dMUDCHAR}*\"
dVALUE		{dDIGIT}+
dSPACE		{dWHITE}+
dARRAY		\(\{
dEARRAY		\}\)
dMAPPING	\(\[
dEMAPPING	\]\)
dCOMMA		,
dCOLON		:
dEOL		\n

%%

{dSPACE}	{ }
{dARRAY}	{ return(MM_ARRAY); }
{dEARRAY}	{ return(MM_EARRAY); }
{dMAPPING}	{ return(MM_MAPPING); }
{dEMAPPING}	{ return(MM_EMAPPING); }
{dSTRING}	{ return(MM_STRING); }
{dVALUE}	{ return(MM_VALUE); }
{dEOL}		{ return(MM_EOL); }
{dCOMMA}	{ return(MM_COMMA); }
{dCOLON}	{ return(MM_COLON); }
<ERROR>.*$	{ BEGIN(INITIAL); return(MM_NONE); }
<*><<EOF>>	{ return(MM_EOF); }
.			{ return(MM_ERR); }

%%

int			 yylval;

int
parse_expect (int vtype)
{
	yylval = yylex();
	if (!(yylval & vtype)) {
		fprintf(stderr, "Syntax error before '%s'\n", yytext);
		fprintf(stderr, "Expected type %d got %d\n", vtype, yylval);
		BEGIN(ERROR);
		return -1;
	}
	return 0;
}

listentry * parse_value_array (void);
listentry * parse_value_mapping (void);

listentry *
parse_value (void)
{
	listentry	*listnode;
	value		*v = NULL;
	char		*cp;

	yylval = yylex();
	switch(yylval) {
		case MM_ARRAY:
			/* Create new array node pointing to array */
			/* parse_value_array() */
			v = value_new(MM_ARRAY);
			v->v_array = (void *)(parse_value_array());
			break;
		case MM_MAPPING:
			v = value_new(MM_MAPPING);
			v->v_array = (void *)(parse_value_mapping());
			break;
		case MM_STRING: /* Create string item */
			while ((cp = strchr(yytext, '\r'))) *cp = '\n';
			v = value_new(MM_STRING);
			v->v_string = strdup(yytext);
			v->v_value = yyleng;
			break;
		case MM_VALUE: /* Create value item */
			v = value_new(MM_VALUE);
			v->v_value = atol(yytext);
			break;

		case MM_COMMA: /* Might as well hande NULLs */
			fprintf(stderr, "Unexpected comma, assuming NULL."
				" before ','\n");
			v = value_new(MM_NONE);
			unput(',');	/* Put the expected comma back */
			break;
		case MM_EOL:
			/* Check stack, terminate arrays */
			break;
		case MM_EOF:
			/* Clear stack, exit cleanly */
			break;
		case MM_EARRAY: /* Got null value again */
			fprintf(stderr, "Unexpected end of array.\n");
			unput(')'); unput('}');
			break;
		case MM_EMAPPING:
			fprintf(stderr, "Unexpected end of mapping\n");
			unput(')'); unput(']');
			break;
		case MM_ERR:
			/* Syntax error, sort this out */
			unput(*yytext);
			BEGIN(ERROR);
			yylval = yylex();
			fprintf(stderr, "Syntax error before '%.8s'", yytext);
			if (yyleng > 8) fprintf(stderr, "...");
			fprintf(stderr, "\n");
			break;
	}

	if (v) {
		listnode = list_item_new();
		listnode->data = (void *)(v);
	} else listnode = NULL;
	return listnode;
}

listentry *
parse_value_mapping (void)
{
	listentry	*root;
	listentry	*node;
	long		 length;
	value		*v;

	root = list_new();
	v = value_new(MM_MAPDATA);

	while(1) {
		node = parse_value();
		if (!node) break;
		// v->v_mapkey = (value *)(node->data);
		list_item_destroy(node, 0);

		yylval = yylex();
		if (yylval != MM_COLON) {
			fprintf(stderr, "Syntax error before '%s'\n", yytext);
			BEGIN(ERROR);
			yylval = yylex();
			return root;
		}

		node = parse_value();
		if (!node) break;
		// v->v_mapdata = (value *)(node->data);
		node->data = (void *)(v);
		list_append(root, node);

		length++;

		yylval = yylex();
		if (yylval == MM_EMAPPING) {
			return root;
		}
		else if (yylval != MM_COMMA) {
			fprintf(stderr, "Syntax error before '%s'\n", yytext);
			BEGIN(ERROR);
			yylval = yylex();
			return root;
		}
	}

	fprintf(stderr, "Syntax error at end of input\n");
	return root;
}

listentry *
parse_value_array (void)
{
	listentry	*root;
	listentry	*node;
	long		 length;

	root = list_new();
	while ((node = parse_value())) {
		length++;
		list_append(root, node);
		yylval = yylex();
		if (yylval == MM_EARRAY) {
			return root;
		}
		else if (yylval != MM_COMMA) {
			fprintf(stderr, "Syntax error before '%s'\n", yytext);
			BEGIN(ERROR);
			yylval = yylex();
			return root;
		}
	}
	fprintf(stderr, "Syntax error at end of input\n");
	return root;
}

int
main (void)
{
	listentry	*tree;
	yyin = stdin;
	yylval = MM_NONE;
	while (!(yylval&(MM_EOF|MM_ERR))) {
		tree = parse_value();
		if (!(yylval&(MM_EOF|MM_EOL|MM_ERR))) parse_expect(MM_EOL);
		if (tree) tree_print(tree);
		else printf("Null input\n");
	}
	return 0;
}
